from typing import Type, Dict, Any, List, Tuple, Set
from pydantic import BaseModel, create_model
import pydantic_resolve.constant as const


def _extract_field_definitions_for_create_model(parent: Type[BaseModel], field_names: List[str]) -> Dict[str, Tuple]:
    """Extract field definitions in format suitable for create_model (Pydantic v2)."""
    field_definitions = {}
    
    for fname in field_names:
        fld = parent.model_fields.get(fname)
        if not fld:
            raise AttributeError(f'field "{fname}" not existed in {parent.__name__}')
        
        # For create_model, we need (type, default_value) tuple
        if fld.is_required():
            field_definitions[fname] = (fld.annotation, ...)
        else:
            # For non-required fields, just use the default
            field_definitions[fname] = (fld.annotation, fld.default)
    
    return field_definitions


def _get_parent_config(parent: Type[BaseModel]) -> Any:
    """Get parent's configuration (Pydantic v2)."""
    return getattr(parent, 'model_config', None)


def _get_parent_validators(parent: Type[BaseModel], field_names: List[str]) -> Dict[str, Any]:
    """Get parent's validators relevant to the subset fields (Pydantic v2)."""
    validators = {}
    
    # In v2, field validators are stored in __pydantic_decorators__
    decorators = getattr(parent, '__pydantic_decorators__', None)
    if decorators and hasattr(decorators, 'field_validators'):
        for validator_name, decorator_info in decorators.field_validators.items():
            # Check if this validator applies to any of our fields
            validator_fields = decorator_info.info.fields
            if any(field in field_names for field in validator_fields):
                # Copy the validator method
                validators[validator_name] = getattr(parent, validator_name)
    
    return validators


def _apply_validators(model_class: Type[BaseModel], validators: Dict[str, Any]) -> None:
    """Apply validators to Pydantic v2 model after creation."""
    # In v2, validators are handled differently and need to be copied manually
    # This is a simplified approach - the proper way would be to use pydantic's internal mechanisms
    for name, validator in validators.items():
        setattr(model_class, name, validator)


def _extract_extra_fields_from_namespace(namespace: Dict[str, Any], disallow: Set[str]) -> Dict[str, Tuple[Any, Any]]:
    """Extract extra field definitions declared on the subset class body.

    Only annotated attributes are considered as Pydantic fields. The result is a
    dict mapping field name -> (annotation, default | Ellipsis) suitable for create_model.

    Args:
        namespace: The class namespace passed into the metaclass with attributes and annotations.
        disallow: Field names that are already selected from the parent and thus cannot be redefined here.

    Raises:
        ValueError: If an extra field duplicates a subset field name.
    """
    annotations: Dict[str, Any] = namespace.get('__annotations__', {}) or {}
    extras: Dict[str, Tuple[Any, Any]] = {}

    for fname, anno in annotations.items():
        if fname == '__pydantic_resolve_subset__':
            continue
        if fname in disallow:
            raise ValueError(f'additional field "{fname}" duplicates subset field')
        default = namespace.get(fname, ...)
        extras[fname] = (anno, default)

    return extras


def _normalize_unique_fields(fields: Any) -> List[str]:
    """Coerce incoming fields into a unique, ordered list of strings.

    Only accepts list or tuple. Other types raise TypeError. All elements must be strings.
    """
    if not isinstance(fields, (list, tuple)):
        raise TypeError('fields must be a list or tuple of field names')
    seq: List[Any] = list(fields)

    for f in seq:
        if not isinstance(f, str):
            raise TypeError('each field name must be a string')

    # Preserve order while removing duplicates
    return list(dict.fromkeys(seq))


def create_subset(parent: Type[BaseModel], fields: List[str], name: str = "SubsetModel") -> Type[BaseModel]:
    """
    Create a subset model from a parent BaseModel using Pydantic's create_model.
    
    Args:
        parent: Parent BaseModel class to create subset from
        fields: List of field names to include in subset
        name: Name of the new subset class (default: "SubsetModel")
        
    Returns:
        A new BaseModel class containing only the specified fields
    """
    if not issubclass(parent, BaseModel):
        raise TypeError('parent must be a pydantic BaseModel')
    
    unique_fields = _normalize_unique_fields(fields)
    field_definitions = _extract_field_definitions_for_create_model(parent, unique_fields)
    validators = _get_parent_validators(parent, unique_fields)

    create_model_kwargs = {}
    config = _get_parent_config(parent)
    if config:
        create_model_kwargs['__config__'] = config
    
    subset_class = create_model(
        name,
        **field_definitions,
        **create_model_kwargs
    )

    _apply_validators(subset_class, validators)
    setattr(subset_class, const.ENSURE_SUBSET_REFERENCE, parent)
    return subset_class


class SubsetMeta(type):
    def __new__(cls, name, bases, namespace, **kwargs):
        subset_info = namespace.get('__pydantic_resolve_subset__')
        if subset_info:
            parent_kls, picked_field = subset_info

            unique_fields = _normalize_unique_fields(picked_field)
            parent_fields = _extract_field_definitions_for_create_model(parent_kls, unique_fields)

            # Then extract extra fields defined in class body
            extra_fields = _extract_extra_fields_from_namespace(namespace, set(unique_fields))

            # Parent validators and config should still apply to subset fields
            create_model_kwargs: Dict[str, Any] = {}
            methods_to_attach: Dict[str, Any] = {}
            attributes_to_attach: Dict[str, Any] = {}

            validators = _get_parent_validators(parent_kls, unique_fields)
            config = _get_parent_config(parent_kls)
            if config:
                create_model_kwargs['__config__'] = config

            # Use the caller's module for better reprs and pickling
            create_model_kwargs['__module__'] = namespace.get('__module__', __name__)

            for nk, nv in namespace.items():
                if nk.startswith('__pydantic_resolve') and nk != '__pydantic_resolve_subset__':
                    attributes_to_attach[nk] = nv
                    continue

                if callable(nv) and (nk.startswith(const.RESOLVE_PREFIX) or nk.startswith(const.POST_PREFIX)):
                    methods_to_attach[nk] = nv


            # Merge field definitions: keep subset order first, then extras
            field_definitions: Dict[str, Tuple[Any, Any]] = {}
            field_definitions.update(parent_fields)
            field_definitions.update(extra_fields)

            print(create_model_kwargs)
            subset_class = create_model(
                name,
                **field_definitions,
                **create_model_kwargs
            )

            _apply_validators(subset_class, validators)

            for method_name, method in methods_to_attach.items():
                setattr(subset_class, method_name, method)
            for attr_name, attr_value in attributes_to_attach.items():
                setattr(subset_class, attr_name, attr_value)
            setattr(subset_class, const.ENSURE_SUBSET_REFERENCE, parent_kls)

            return subset_class
            
        # Allow defining the base marker class without warning, bypass
        if name == 'DefineSubset' and not bases:
            return super().__new__(cls, name, bases, namespace, **kwargs)

        raise ValueError(f'Class {name} must define __pydantic_resolve_subset__ to use SubsetMeta')

class DefineSubset(metaclass=SubsetMeta):
    pass



from dataclasses import dataclass
from typing import Type, Any, Callable, Optional
from pydantic import BaseModel, model_validator, Field
import warnings

import pydantic_resolve.constant as const
from pydantic_resolve.utils import class_util
from pydantic_resolve.utils.depend import LoaderDepend


class Relationship(BaseModel):
    field: str  # fk name

    # use biz to distinguish multiple same target_kls under same field
    biz: Optional[str] = Field(default=None, min_length=1)  # Optional or non-empty string
    target_kls: Any

    # default value when fk is None.
    # Two params are exclusive. A user may explicitly set default_none_val=None.
    # We must distinguish "not provided" vs "provided with value None".
    field_none_default: Optional[Any] = None
    field_none_default_factory: Optional[Callable[[], Any]] = None

    loader: Optional[Callable]

    @model_validator(mode="after")
    def _validate_defaults(self) -> "Relationship":
        # Avoid evaluating the deprecated fallback unless necessary to prevent warnings.
        fields_set = getattr(self, 'model_fields_set', set())
        val_set = 'field_none_default' in fields_set
        factory_set = 'field_none_default_factory' in fields_set
        if val_set and factory_set:
            raise ValueError(
                "field_none_default and field_none_default_factory cannot both be defined"
            )
        return self

class ErConfig(BaseModel):
    kls: Type[BaseModel]
    relationships: list[Relationship]

    @model_validator(mode="after")
    def _validate_relationships(self) -> "ErConfig":
        rels = self.relationships or []

        # helper to make potentially unhashable target_kls hashable for set/dict keys
        def _hashable(x: Any) -> Any:
            try:
                hash(x)
                return x
            except TypeError:
                return repr(x)

        # Disallow duplicate (field, biz, target_kls) triples
        seen = set()
        for r in rels:
            key = (r.field, r.biz, _hashable(r.target_kls))
            if key in seen:
                raise ValueError(
                    f"Duplicate relationship detected for (field={r.field!r}, biz={r.biz!r}, target_kls={r.target_kls!r})"
                )
            seen.add(key)

        return self

class ErDiagram(BaseModel):
    configs: list[ErConfig]

    @model_validator(mode="after")
    def _validate_configs(self) -> "ErDiagram":
        cfgs = self.configs or []
        seen = set()
        for cfg in cfgs:
            kls = cfg.kls
            if kls in seen:
                raise ValueError(f"Duplicate config.kls detected: {kls}")
            seen.add(kls)
        return self

    description: Optional[str] = None


@dataclass
class LoaderInfo:
    field: str
    biz: Optional[str] = None


def LoadBy(key: str, biz: Optional[str] = None) -> LoaderInfo:
    return LoaderInfo(field=key, biz=biz)


class ErPreGenerator:
    def __init__(self, er_diagram: Optional[ErDiagram]) -> None:
        self.er_configs_map = {config.kls: config for config in er_diagram.configs} if er_diagram else None

    def _identify_config(self, target: Type) -> ErConfig:
        """Locate the matching ErConfig for a target class via compatibility check."""
        for kls, cfg in self.er_configs_map.items():
            if class_util.is_compatible_type(target, kls):
                return cfg
        raise AttributeError(f'No ErConfig found for {target.__name__}')

    def _identify_relationship(self, config: ErConfig, loadby: str, biz: Optional[str], target_kls: Type) -> Relationship:
        """Find the relationship matching (field=loadby, biz, target_kls)."""
        for rel in config.relationships:
            if rel.field != loadby:
                continue
            if class_util.is_compatible_type(target_kls, rel.target_kls) and biz == rel.biz:
                return rel
        raise AttributeError(
            f'Relationship for "{target_kls.__name__}" using "{loadby}" not found'
        )

    def prepare(self, kls: Type):
        """Auto-generate resolve_XXX methods for fields annotated with LoadBy metadata.

        For each pydantic field carrying LoadBy, create a resolve method that uses the
        corresponding relationship's loader via LoaderDepend.
        """
        auto_loader_fields = list(class_util.get_pydantic_field_items_with_load_by(kls))

        if not auto_loader_fields:
            return 

        if self.er_configs_map is None:
            raise ValueError('er_configs_map is None, cannot identify config')

        config = self._identify_config(kls)

        for field_name, loader_info, annotation in auto_loader_fields:
            method_name = f'{const.RESOLVE_PREFIX}{field_name}'
            if hasattr(kls, method_name):
                warnings.warn(
                    f'{method_name} already exists in {kls.__name__}, skipping auto-generation.'
                )
                continue

            relationship = self._identify_relationship(
                config=config,
                loadby=loader_info.field,
                biz=loader_info.biz,
                target_kls=annotation,
            )

            if relationship.loader is None:
                # loader is optional in Relationship, but required for auto-generation
                raise AttributeError(f'Loader not provided in relationship for field "{loader_info.field}" in class "{kls.__name__}"')

            def create_resolve_method(key: str, rel: Relationship):  # closure per field
                def resolve_method(self, loader=LoaderDepend(rel.loader)):
                    fk = getattr(self, key)
                    if fk is None:
                        fields_set = getattr(rel, 'model_fields_set', set())
                        if 'field_none_default' in fields_set:
                            return rel.field_none_default  # may be None intentionally

                        if rel.field_none_default_factory is not None:
                            return rel.field_none_default_factory()
                        return None
                    return loader.load(fk)
                resolve_method.__name__ = method_name
                resolve_method.__qualname__ = f'{kls.__name__}.{method_name}'
                return resolve_method

            setattr(kls, method_name, create_resolve_method(loader_info.field, relationship),)
